/*
 * Description:  
 * Comments:
 *           The Neopixel library can be found here together with examples
 *           https://github.com/adafruit/Adafruit_NeoPixel
 *           
 * 2017-12-25: First release of the software.
 * 2018-02-08: Final release of 1. software version.
 * 2018-10-16: Updated WEB interface. Updated EEPROM handling.
 * 2019-12-09: Updated WiFi and NEOpixel handling.
 * 2019-12-26: All new WEB interface. Uses SPIFFS to hold HTML, JS and CSS files
 *
 * When fully configured and connected to the WiFi enter the WEB interface by entering the following in a browser:
 * http//<IP address>/NEOPixel_new_20191222.html
 *
 */
 
/*----|| INCLUDES ||--------------------------------------------*/ 
#include <EEPROM.h>
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <Adafruit_NeoPixel.h>
#include <FS.h>   // Include the SPIFFS library
#ifdef __AVR__
  #include <avr/power.h>
#endif
/*----|| INCLUDES-end ||----------------------------------------*/ 

/*----|| #DEFINES ||--------------------------------------------*/ 
#define PIN     D8 // Connected to Neopixel IN.
#define NUM_LEDS 7 // Change to fit actual LEDS.
#define DEBUG    1 // 0: No debug info. 1: Show debug info.

#define L_STATUS      1 // Single byte variable
#define C_RED         2 // Single byte variable
#define C_GREEN       3 // Single byte variable
#define C_BLUE        4 // Single byte variable
#define BRIGHTNESS    5 // Single byte variable
#define DELAY         6 // Single byte variable 
#define RAINBOWCYCLE  7 // Single byte variable 
#define RAINBOW       8 // Single byte variable
#define STATIC        9 // Single byte variable 
#define FADE          9 // Single byte variable
#define CANDLE       10 // Single byte variable
#define SCEENE1      11 // Dummy for additional sceenes
#define SCEENE2      12 // Dummy for additional sceenes
#define SCEENE3      13 // Dummy for additional sceenes
#define SSID_POS     15 
#define SSID_LEN     21 // Max 21 char incl 0x00
#define PASSWORD_POS 40
#define PASSWORD_LEN 21 // Max 21 char incl 0x00
/*----|| #DEFINES-end ||----------------------------------------*/

/*----|| VARIABLES ||--------------------------------------------*/
// Taken from the Neopixel Library.
// Parameter 1 = number of pixels in strip
// Parameter 2 = Arduino pin number (most are valid)
// Parameter 3 = pixel type flags, add together as needed:
//   NEO_KHZ800  800 KHz bit stream (most NeoPixel products w/WS2812 LEDs)
//   NEO_KHZ400  400 KHz (classic 'v1' (not v2) FLORA pixels, WS2811 drivers)
//   NEO_GRB     Pixels are wired for GRB bit stream (most NeoPixel products)
//   NEO_RGB     Pixels are wired for RGB bit stream (v1 FLORA pixels, not v2)
//   NEO_RGBW    Pixels are wired for RGBW bit stream (NeoPixel RGBW products)
Adafruit_NeoPixel strip = Adafruit_NeoPixel(NUM_LEDS, PIN, NEO_GRB + NEO_KHZ800);

String version = "\nIKEA Hack server ver: 2019-12-21 V:1.0"; 

/* ---- Network variables --------- */
char ssid[21]     = "ssid";   // Your WiFi networks SSID.
char password[21] = "password"; // Password to your WiFi network.
ESP8266WebServer server(80);
String HTMLMainFile = "/NEOPixel_new_20191222.html"; // The main HTML file to load
/* ---- Network variables END --------- */

/* Configuration variables ------------ */
byte lampStatus       = 1;  // 0 or 1.
byte lampRed          = 255;// 0 - 255
byte lampGreen        = 0;  // 0 - 255
byte lampBlue         = 0;  // 0 - 255
byte lampBrightness   = 255;// 0 - 255
byte lampDelay        = 20; // 0 - 255
byte lampRainbowCycle = 0;  // 0 or 1.
byte lampRainbow      = 1;  // 0 or 1.
byte lampStatic       = 0;  // 0 or 1.
byte lampFade         = 0;  // 0 or 1.
byte lampCandle       = 0;  // 0 or 1.
// Add additinal sceene variables here.
/* Configuration variables END ------- */

byte loopBreak = 0; // Break out of a illumination loop
byte handleApi = 0; // 1: Received a HTTP POST request via API interface. 

// See the description in the documentation
byte fadeInterpolation[97] =
{ 6,  6,  6,  7,  7,  8,  9, 10, 10, 11, //  1
 12, 13, 13, 14, 15, 16, 17, 19, 20, 22, //  2
 23, 25, 26, 28, 29, 32, 35, 38, 40, 45, //  3
 49, 53, 57, 62, 66, 71, 75, 79, 82, 86, //  4
 89, 91, 93, 95, 97, 98, 99,100,100,100, //  5
 99, 98, 97, 95, 93, 91, 89, 86, 82, 79, //  6
 75, 71, 66, 62, 57, 53, 49, 45, 40, 38, //  7
 35, 32, 29, 28, 26, 25, 23, 22, 20, 19, //  8
 17, 16, 15, 14, 13, 13, 12, 11, 10, 10, //  9
  9,  8,  7,  7,  6,  6,  6 };           // 10 (97)
  
// Setup variable
const int     bSize = 22;
      char    Buffer[bSize]; // Serial buffer
      int     ByteCount;
      String  inData;
      int     inByte = 0;    // incoming serial byte
      byte    count  = 0;
      boolean cont;          // Continue looking for setup commands
      boolean setupNotDone;  // Setup done start lamp.
/*----|| VARIABLES ||--------------------------------------------*/

/*----|| FUNCTIONS ||--------------------------------------------*/
/*---------------------------------------------------------------*/
/*  Utility function                                             */
/*  Visual error indication via LED at pin D2                    */
/*  This is a dead end. You will need to fix the error and reset */
/*    the device                                                 */   
/*  State 1: No WiFi connection available.                       */
/*---------------------------------------------------------------*/
void errorState(int state)
{
   int i = 0;
   Serial.println("Entered error state.");
   if( state == 1) // NO WiFi available
   {
     while(1)
     {
       strip.setPixelColor(1, strip.Color(128, 0, 0) ); strip.show(); // Visual indication for Error
       Serial.print(".");
       delay(500);
       strip.setPixelColor(1, strip.Color(0, 0, 0) ); strip.show();       
       delay(300);   
       if( i++ > 60 )
       {
         i = 0;
         Serial.println("Entered error state 1. Press reset");
       }
     }          
   }
}

/*---------------------------------------------------------------*/
/* Utility function                                              */
/* Halt the program and wait for "key press" at the serial port. */
/* For DEBUG purpose only                                        */
/*---------------------------------------------------------------*/
void keypressWait()
{
  Serial.println("** Hit a key to continue");
  while(Serial.available() == 0){}
  while(Serial.available()){Serial.read();}
}

/*---------------------------------------------------------------*/
/* Checks if a string is numeric.                                */
/* Removed check for decimal point as it was not needed          */
/* Original code found here:                                     */
/*     http://tripsintech.com/arduino-isnumeric-function/        */
/*---------------------------------------------------------------*/
boolean isNumeric(String str) 
{
  unsigned int stringLength = str.length();
 
  if (stringLength == 0) return false;
 
  for(unsigned int i = 0; i < stringLength; ++i) 
  {
     if (isDigit(str.charAt(i))) 
     {
       continue;
     }
     return false;
  }
  return true;
}

/*--------------------------------------------------------------------*/
/*  Utility function                                                  */
/*  Original code found here:                                         */
/*  https://tttapa.github.io/ESP8266/Chap16%20-%20Data%20Logging.html */
/*------------------------------------------------------------------- */
String formatBytes(size_t bytes)  // convert sizes in bytes to KB and MB
{
  if (bytes < 1024) 
  {
    return String(bytes) + "B";
  } 
  else if (bytes < (1024 * 1024)) 
  {
    return String(bytes / 1024.0) + "KB";
  } 
  else if (bytes < (1024 * 1024 * 1024)) 
  {
    return String(bytes / 1024.0 / 1024.0) + "MB";
  }
}

/*---------------------------------------------------------------*/
/* Start the SPI Flash File System (SPIFFS)                      */
/*---------------------------------------------------------------*/
int startSPIFFS(void)
{
  if( WiFi.status() != WL_CONNECTED )
  {
    // There is no WiFi connection. Reconnect.
    if( DEBUG ) Serial.println("** ERROR: No WiFi available");
    // Try and reconnect WiFi
    if ( !doWifiConnect() ) return false; // Could not reconnect !
  }
 
  // Start the SPIFFS and list all contents
  Serial.print("** Starting SPIFFS.");
  SPIFFS.begin();

  Serial.println(" [OK]");
  Serial.println("** Contents:");
  {
    Dir dir = SPIFFS.openDir("/");
    while (dir.next())                       // List the file system contents
    {
      String fileName = dir.fileName();
      size_t fileSize = dir.fileSize();
      Serial.printf("\t** FS File: %s, size: %s\r\n", fileName.c_str(), formatBytes(fileSize).c_str());
    }
    Serial.printf("\n");
  }
  return true;
}

/*--------------------------------------------------------------------*/
/* Original code found here:                                          */
/*  https://tttapa.github.io/ESP8266/Chap16%20-%20Data%20Logging.html */
/*--------------------------------------------------------------------*/
String getContentType(String filename) 
{ // convert the file extension to the MIME type
  if (filename.endsWith(".html")) return "text/html";
  else if (filename.endsWith(".css")) return "text/css";
  else if (filename.endsWith(".js")) return "application/javascript";
  else if (filename.endsWith(".ico")) return "image/x-icon";
  else if (filename.endsWith(".svg")) return "image/svg+xml"; // HHS new 2019-12-18
  return "text/plain";
}

/*---------------------------------------------------------------*/
/* Holds the file: "NEOPixel_new_20191222_load.js"               */
/* This file includes the EEPROM stored values, which is needed  */
/*  in order to update the WEB configuration interface after     */
/*  reset or power up of the lamp.                               */
/*---------------------------------------------------------------*/
void NEOPixel_load_script ()
{
  Serial.println("Received request");
  Serial.println("  handleFileRead: NEOPixel_load_script.");
  
  String lampON = "OFF";
  String lampChecked = "false";
  if (lampStatus ) 
  {
	  lampON = "ON";
	  lampChecked = "true";
  }
  String sceene = "";
  String colour = "";
  if( lampRainbowCycle ) sceene = "Rainbowcycle is active.";
  else if( lampRainbow ) sceene = "Rainbow is active.";
  else if( lampCandle ) sceene = "Candle is active.";
  else if( lampStatic ) colour = "Static is active.";
  else if( lampFade ) colour = "Fade is active.";
  else sceene = "Unkown is active.";
  String temp  = "document.addEventListener('DOMContentLoaded', function ()\n";
         temp += "{\n";
         temp += "  document.getElementById(\"activeSceene\").innerHTML = \"";
         temp += sceene;
         temp += "\";\n";
         temp += "  document.getElementById(\"activeButton\").innerHTML = \"";
         temp += colour;
         temp += "\";\n";
         temp += "  document.getElementById(\"onOffButton\").checked = \"";
         temp += lampChecked;
         temp += "\";\n";
         temp += "  document.getElementById(\"lampStatus\").innerHTML = \"";
         temp += lampON;
         temp += "\";\n";
         temp += "  document.getElementById(\"brightnessSlider\").value = \"";
         temp += lampBrightness;
         temp += "\";\n";
         temp += "  document.getElementById(\"brightnessValue\").innerHTML = \"";
         temp += lampBrightness;
         temp += "\";\n";
         temp += "  document.getElementById(\"delaySlider\").value = \"";
         temp += lampDelay;
         temp += "\";\n";
         temp += "  document.getElementById(\"delayValue\").innerHTML = \"";
         temp += lampDelay;
         temp += "\";\n";
         temp += "  document.getElementById(\"redSlider\").value = \"";
         temp += lampRed;
         temp += "\";\n";
         temp += "  document.getElementById(\"redValue\").innerHTML = \"";
         temp += lampRed;
         temp += "\";\n";
         temp += "  document.getElementById(\"greenSlider\").value = \"";
         temp += lampGreen;
         temp += "\";\n";
         temp += "  document.getElementById(\"greenValue\").innerHTML = \"";
         temp += lampGreen;
         temp += "\";\n";
         temp += "  document.getElementById(\"blueSlider\").value = \"";
         temp += lampBlue;
         temp += "\";\n";
         temp += "  document.getElementById(\"blueValue\").innerHTML = \"";
         temp += lampBlue;
         temp += "\";\n";
         temp += "}, false);\n";
		 
  server.send ( 200, "application/javascript", temp );
  Serial.println("  NEOPixel_load_script transmitted.");
} // String getPage ()

/*---------------------------------------------------------------*/
/* This function is called when a HTTP POST request is received  */
/*   via http://<ip address>/API                                 */
/*---------------------------------------------------------------*/
void handleAPI() 
{
  handleApi = 1;
  handleRoot();
} // void handleAPI() END 

/*---------------------------------------------------------------*/
/* This function reads the incoming arguments and updates the   */
/* status of the lamp.                                           */
/* http://<ip address>/                                          */
/*---------------------------------------------------------------*/
void handleRoot() 
{
  String str, strText;

  if( DEBUG ) Serial.println("** HandleRoot called.");    
  loopBreak = 1; // If called from loop, break out of it.
  
  if (server.hasArg("LAMP") )
  {
    str = server.arg("LAMP");
    if ( str == "1" ) 
    {
      // Received LAMP=1
      lampStatus = 1;
      EEPROM.write(L_STATUS, 1); // Save new lamp status
      EEPROM.commit();           // Needed in order to save content to flash.
      if( DEBUG ) Serial.println("**  LAMP ON.");
      if( handleApi )
      {  // Request received via API
         server.send ( 200, "text/html", "OK" ); // Return OK
      }

      if( DEBUG ) Serial.println("**  New lamp status saved (ON).");
    } 
    else if ( str == "0" ) 
    {
      // Received LAMP=0
      lampStatus = 0;
      EEPROM.write(L_STATUS, 0); // Save new lamp status
      EEPROM.commit();           // Needed in order to save content to flash.
      if( DEBUG ) Serial.println("**  LAMP OFF.");
      if( handleApi )
      {  // Request received via API
         server.send ( 200, "text/html", "OK" ); // Return OK
      }

      for(uint16_t i = 0; i < strip.numPixels(); i++) 
      {
        strip.setPixelColor(i, strip.Color(0, 0, 0) );
        strip.show();
      }      
      if( DEBUG ) Serial.println("**  New lamp status saved (OFF).");      
    } 
    else 
    {
      // Received something wrong.
      Serial.println("Err Lamp status.");
      if( handleApi )
      {  // Request received via API
         server.send ( 200, "text/html", "ERROR" ); // Return OK
      }
    } 
    if( DEBUG ) Serial.println("**  LAMP submit handled.");    
  }
  else if( server.hasArg("RED") || server.hasArg("GREEN") || server.hasArg("BLUE") )
  { 
    boolean colourError = 0;
    strText = "COLOUR: ";
    if( server.hasArg("RED") )
    {
      // Received RED=<value>
      str = server.arg("RED");
      strText += "RED: ";      
      if( isNumeric( str ) && (str.toInt() < 256) )
      {
        if( DEBUG ) Serial.println("**  Data received are numeric");
        lampRed  = str.toInt();
        strText += str;
        if( DEBUG ) 
        {
          Serial.print("**  New RED value set: "); Serial.println(lampRed);
        }
      }
      else
      {
        if( DEBUG ) Serial.println("**  Data received are NOT numeric");
        colourError = 1;
        strText += "N/A ";
        if( DEBUG ) 
        {
          Serial.print("**  Error new RED value: "); Serial.println(str);
        }
      }
    }
    if(server.hasArg("GREEN") )
    {
      // Received GREEN=<value>
      str = server.arg("GREEN");
      strText += " GREEN: ";
      if( isNumeric( str ) && (str.toInt() < 256) )
      {
        if( DEBUG ) Serial.println("**  Data received are numeric");
        lampGreen = str.toInt();
        strText  += str;
        if( DEBUG ) 
        {
          Serial.print("**  New GREEN value set: "); Serial.println(lampGreen);
        }
      }
      else
      {
        if( DEBUG ) Serial.println("**  Data received are NOT numeric");
        colourError = 1;
        strText += "N/A ";
        if( DEBUG ) 
        {
          Serial.print("**  Error new GREEN value: "); Serial.println(str);
        }
      }
    }
    if(server.hasArg("BLUE") )
    {
      // Received BLUE=<value>
      str      = server.arg("BLUE");
      strText += " BLUE: ";
      if( isNumeric( str ) && (str.toInt() < 256) )
      {
        if( DEBUG ) Serial.println("**  Data received are numeric");
        lampBlue = str.toInt();
        strText += str;
        if( DEBUG ) 
        {
          Serial.print("**  New BLUE value set: "); Serial.println(lampBlue);
        }
      }
      else
      {
        if( DEBUG ) Serial.println("**  Data received are NOT numeric");
        colourError = 1;
        strText    += "N/A ";
        if( DEBUG ) 
        {
          Serial.print("**  Error new BLUE value: "); Serial.println(str);
        }
      }
    } 
    if( handleApi )
    {  // Request received via API
      if( colourError )
        server.send ( 200, "text/html", "ERROR" ); // Return ERROR
      else
        server.send ( 200, "text/html", "OK" ); // Return OK
    }

    if( DEBUG ) Serial.println("**  COLOUR submit handled.");  
      
    for( uint16_t i = 0; i < strip.numPixels(); i++) 
    {
      strip.setPixelColor(i, strip.Color(lampRed, lampGreen, lampBlue) );
      strip.show();
    }
  }
  else if (server.hasArg("SCEENE") )
  {
    if( DEBUG ) Serial.println("** Received SCEENE."); 
    str = server.arg("SCEENE");
    boolean accepted = 1;
	  if( str == "RAINBOWCYCLE" )
	  {
      // Received SCEENE=RAINBOWCYCLE
      if( DEBUG ) Serial.println("**  Rainbowcycle");
      lampRainbowCycle = 1;
      lampRainbow = lampStatic  = lampFade = lampCandle = 0; 
	  }
	  else if( str == "RAINBOW" )
	  {
      // Received SCEENE=RAINBOW
      if( DEBUG ) Serial.println("**  Rainbow");
      lampRainbowCycle = 0;
      lampRainbow = 1;
	    lampStatic  = lampFade = lampCandle = 0; 
	  }
	  else if( str == "CANDLE" )
	  {
      // Received SCEENE=CANDLE
      if( DEBUG ) Serial.println("**  Candle");
      lampRainbowCycle = lampRainbow = lampStatic  = lampFade = 0;
	    lampCandle = 1; 
	  }
	  else
  	{
      accepted = 0;
    }
    if( handleApi )
    {  // Request received via API
      if( accepted )
        server.send ( 200, "text/html", "OK" ); // Return OK
      else
        server.send ( 200, "text/html", "ERROR" ); // Return ERROR
    }
    else
    {   
      if( accepted )
        strText = "SCEENE received correct";
      else
        strText = "Incorrect SCEENE received";
      handleFileRead( HTMLMainFile ); // Update WEB interface
    }
    if( DEBUG ) Serial.println("**  SCEENE submit handled."); 
  }
  else if(server.hasArg("BRIGHTNESS") )
  {
    // Received BRIGHTNESS=<value>
    str = server.arg("BRIGHTNESS");
    // Check if received value is numeric AND below 256
    if( isNumeric( str ) && (str.toInt() < 256) )
    {
      if( DEBUG ) Serial.println("**  Data received are numeric");
      lampBrightness = str.toInt();
      if( handleApi )
      {
        server.send ( 200, "text/html", "OK" ); // Return OK  
      }
      else
      {
        handleFileRead( HTMLMainFile ); // Update WEB interface
      }
      strip.setBrightness(lampBrightness);
      strip.show();
      if( DEBUG )
      {
        Serial.print("**  New BRIGHTNESS value set: "); Serial.println(lampBrightness);
      }          
    }
    else
    {
      if( DEBUG ) Serial.println("**  Data received are NOT numeric");
      if( handleApi )
      {
        server.send ( 200, "text/html", "ERROR" ); // Return ERROR  
      }
      else
      {
        strText += "BRIGHTNESS error Wrong data: ";
        strText += str;
        handleFileRead( HTMLMainFile ); // Update WEB interface
      }
    }  
  }
  else if(server.hasArg("DELAY") )
  {
    // Received DELAY=<value>
    str = server.arg("DELAY");
    // Check if received value is numeric AND below 256
    if( isNumeric( str ) && (str.toInt() < 256) )
    {
      if( DEBUG ) Serial.println("**  Data received are numeric");
      lampDelay = str.toInt();
      if( handleApi )
      {
        server.send ( 200, "text/html", "OK" ); // Return OK  
      }
      else
      {
        handleFileRead( HTMLMainFile ); // Update WEB interface
      }
      if( DEBUG )
      {
        Serial.print("**  New DELAY value set: "); Serial.println(lampDelay);
      }          
    }
    else
    {
      if( DEBUG ) Serial.println("**  Data received are NOT numeric");
      if( handleApi )
      {
        server.send ( 200, "text/html", "ERROR" ); // Return ERROR  
      }
      else
      {
        handleFileRead( HTMLMainFile ); // Update WEB interface
      }
    }  
  }
  else if(server.hasArg("COLOUR"))
  {
	str = server.arg("COLOUR");
    boolean accepted = 1;
	if( str == "FADE" )
	{
      // Received COLOUR=FADE
      if( DEBUG ) Serial.println("**  Fade");
      lampRainbowCycle = lampRainbow = lampStatic  = 0;
	  lampFade = 1;
	  lampCandle = 0;
	}
	else if( str == "STATIC" )
	{
    // Received COLOUR=STATIC
    if( DEBUG ) Serial.println("**  Static");
    lampRainbowCycle = lampRainbow = 0;
    lampStatic  = 1;
	  lampFade = lampCandle = 0;
	}
	else
	{
      accepted = 0;
	}
    if( handleApi )
    {  // Request received via API
      if( accepted )
        server.send ( 200, "text/html", "OK" ); // Return OK
      else
        server.send ( 200, "text/html", "ERROR" ); // Return ERROR
    }
    else
    {   
      if( accepted )
        strText = "COLOUR received correct";
      else
        strText = "Incorrect COLOUR received";
      handleFileRead( HTMLMainFile ); // Update WEB interface
    }
    if( DEBUG ) Serial.println("**  COLOUR submit handled."); 
  }
  else if(server.hasArg("SAVE") )
  {
    // Save present configuration in EEPROM.
    EEPROM.write(L_STATUS,     lampStatus);
    EEPROM.write(C_RED,        lampRed);
    EEPROM.write(C_GREEN,      lampGreen);
    EEPROM.write(C_BLUE,       lampBlue);
    EEPROM.write(BRIGHTNESS,   lampBrightness);
    EEPROM.write(DELAY,        lampDelay);	
    EEPROM.write(RAINBOWCYCLE, lampRainbowCycle);	
    EEPROM.write(RAINBOW,      lampRainbow);
    EEPROM.write(STATIC,       lampStatic);
    EEPROM.write(FADE,         lampFade);
    EEPROM.write(CANDLE,       lampCandle);
    // Comment: We dont save the SSID or Password to the WiFi as it is already saved in EEPROM.
    EEPROM.commit(); // Needed in order to save content to flash.
    
    if( handleApi )
    {  // Request received via API
       server.send ( 200, "text/html", "OK" ); // Return OK
    }

    if( DEBUG ) Serial.println("**  New configuration saved.");    
  }
  else
  {
    if( handleApi )
    {  // Request received via API
       server.send ( 200, "text/html", "ERROR" ); // Return ERROR
    }
    
    if( DEBUG ) Serial.println("**  html text send.");
  }
  handleApi = 0;
} // void handleRoot() END

/*--------------------------------------------------------------------*/
/*                                                                    */
/*--------------------------------------------------------------------*/
void handleNotFound(void)  // if the requested file or page doesn't exist, return a 404 not found error
{
  Serial.println("Received request");
  if (!handleFileRead(server.uri())) // check if the file exists in the flash memory (SPIFFS), if so, send it
  {
    server.send(404, "text/plain", "404: File Not Found");
  }
}
/*--------------------------------------------------------------------*
 * Original code found here:                                          *
 *  https://tttapa.github.io/ESP8266/Chap16%20-%20Data%20Logging.html *
 *--------------------------------------------------------------------*/
bool handleFileRead(String path) 
{ 
  // send the right file to the client (if it exists)
  Serial.println("  handleFileRead: " + path);
  if (path.endsWith("/")) path += "index.html";         // If a folder is requested, send the index file
  String contentType = getContentType(path);            // Get the MIME type
  if (SPIFFS.exists(path))                              // If the file exists
  {                            
    File file = SPIFFS.open(path, "r");                 // Open it
    size_t sent = server.streamFile(file, contentType); // And send it to the client
    file.close();                                       // Then close the file again
    return true;
  }
  Serial.println("\tFile Not Found");
  return false;                                         // If the file doesn't exist, return false
}

/*---------------------------------------------------------------*/
/* Fill the LEDs one after the other with a color                */
/*---------------------------------------------------------------*/
void colorWipe(uint32_t c, uint8_t wait) 
{
  uint16_t i;
  for( i = 0; i < strip.numPixels(); i++) 
  {
    strip.setPixelColor(i, c);
    strip.show();

    server.handleClient(); // Check for incoming requests
    if( loopBreak )
    {
      loopBreak = 0;
      return;
    }
    delay(wait);
  }
} // void colorWipe() END 

/*---------------------------------------------------------------*/
/* All 7 LEDs will show the same colour.                         */
/* The 7 LEDs will slowly change through a Rainbow of colours.   */
/* Speed adjusted by variable wait                               */ 
/* Reuse from NEOpixel example.                                  */
/*---------------------------------------------------------------*/
void rainbow(uint8_t wait) 
{
  uint16_t i, j;

  for( j = 0; j < 256; j++) 
  {
    for( i = 0; i < strip.numPixels(); i++) 
    {
      strip.setPixelColor(i, Wheel((i+j) & 255));
    }
    strip.show();
    server.handleClient(); // Check for incoming requests
    if( loopBreak )
    {
      loopBreak = 0;
      return;
    }
    delay(wait);
  }
} // void rainbow() END 

/*---------------------------------------------------------------*/
/* The 6 LEDs in the outer ring will cycle through a Rainbow of  */ 
/* colours. The center LED will be turned off.                   */
/* Speed adjusted by variable wait                               */
/* Taken from NEOpixel example.                                  */ 
/*---------------------------------------------------------------*/
void rainbowCycle(uint8_t wait) 
{
  uint16_t i, j;

  strip.setPixelColor(0, 0, 0, 0);
  
  for(j=0; j<256*5; j++) 
  { // 5 cycles of all colors on wheel
    for(i=1; i< strip.numPixels(); i++) 
    {
      strip.setPixelColor(i, Wheel(((i * 256 / strip.numPixels()) + j) & 255));
    }
    strip.show();
    server.handleClient(); // Check for incoming requests
    if( loopBreak )
    {
      loopBreak = 0;
      return;
    }
    delay(wait);
  }
} // void rainbowCycle() END

/*---------------------------------------------------------------*/
/* Input a value 0 to 255 to get a color value.                  */
/* The colours are a transition r - g - b - back to r.           */
/* Used by Rainbow and RainbowCycle function                     */
/*---------------------------------------------------------------*/ 
uint32_t Wheel(byte WheelPos) 
{
  WheelPos = 255 - WheelPos;
  if(WheelPos < 85) 
  {
    return strip.Color(255 - WheelPos * 3, 0, WheelPos * 3);
  }
  if(WheelPos < 170) 
  {
    WheelPos -= 85;
    return strip.Color(0, WheelPos * 3, 255 - WheelPos * 3);
  }
  WheelPos -= 170;
  return strip.Color(WheelPos * 3, 255 - WheelPos * 3, 0);
}

/*---------------------------------------------------------------*/
/* Will fade the colour set for static from off to full on       */
/* Fade curve defined by array fadeInterpolation[]               */
/*---------------------------------------------------------------*/
void fade(uint8_t wait)
{
  byte r,g,b;
  byte i,j; 

  for( i = 0; i < 97; i++)
  {
    r = (byte)( (float)lampRed   * (float)fadeInterpolation[i] / (float)100.0);
    g = (byte)( (float)lampGreen * (float)fadeInterpolation[i] / (float)100.0);
    b = (byte)( (float)lampBlue  * (float)fadeInterpolation[i] / (float)100.0);

    for( j = 0; j < strip.numPixels(); j++) 
    {
      strip.setPixelColor(j, r, g, b);
      strip.show();      
      delay(wait);
    }  
    server.handleClient(); // Check for incomming requests
    if( loopBreak )
    {
      loopBreak = 0;
      return;
    }
  }
} // void fade() END

/*---------------------------------------------------------------*/
/* Test function simulation of a candle (need improvement)       */
/* https://moving-rainbow.readthedocs.io/en/latest/guide/candle-flicker/ */
/*---------------------------------------------------------------*/
void candle(uint8_t wait) 
{
   uint8_t green; // brightness of the green 
   uint8_t red;   // add a bit for red
   for(uint8_t i=0; i<100; i++) 
   {
     green = 30 + random(155);
     red = green + random(70);
     strip.setPixelColor(random(strip.numPixels() - 1), red, green, 0);
     strip.show();
     delay(wait);
     server.handleClient(); // Check for incomming requests
     if( loopBreak )
     {
       loopBreak = 0;
       return;
     }     
   }
}

/*---------------------------------------------------------------*/
/* Test function. Displays random colours                        */
/*---------------------------------------------------------------*/
void colourRandum(uint8_t wait)
{
  byte i, r, g, b;
  while ( 1 )
  {
    r = random(255);
    g = random(255);
    b = random(255);

    for( i = 0; i < strip.numPixels(); i++) 
    {
      strip.setPixelColor(i, r, g, b);
      strip.show();      
    }
    strip.show();
  
    server.handleClient(); // Check for incoming requests
    if( loopBreak )
    {
      loopBreak = 0;
      return;
    }
    delay(wait);
  }
}

/*---------------------------------------------------------------*/
/*  Utility function                                             */
/*  Software found here:                                         */
/*   https://github.com/esp8266/Arduino/issues/4060              */  
/*
/*---------------------------------------------------------------*/
bool doWifiConnect() 
{
  if( DEBUG ) Serial.println("** Reconnecting WLAN...");
  Serial.print("   ** WiFi.status: "); showWiFiStatus( WiFi.status() );
  //WiFi.disconnect(true);
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);

  int waitCounter = 0;
  while (WiFi.status() != WL_CONNECTED) 
  {
    waitCounter++;
    if (waitCounter == 30)
    {
      Serial.println("** ERROR: could not reconnect WiFi");
      return false;
    }
    delay(500);
  }
  if( DEBUG ) 
  {
    Serial.print("   ** waitCounter: "); Serial.println(waitCounter);
  }  
  Serial.print("** WiFi reconnected with status: "); showWiFiStatus( WiFi.status() );
  return true;
}

/*---------------------------------------------------------------*/
/*  Utility function                                             */
/*  Display WiFi status in clear text                            */
/*  Input: # indicating the WiFi status to show.                 */
/*---------------------------------------------------------------*/
void showWiFiStatus( int i )
{
   switch (i)
   {
      case 255:
        Serial.println("WL_NO_SHIELD");
        break;
      case 0:
        Serial.println("WL_IDLE_STATUS");
        break;
      case 1:
        Serial.println("WL_NO_SSID_AVAIL");
        break;
      case 2:
        Serial.println("WL_SCAN_COMPLETED");
        break;
      case 3:
        Serial.println("WL_CONNECTED");
        break;
      case 4:
        Serial.println("WL_CONNECT_FAILED");
        break;
      case 5:
        Serial.println("WL_CONNECTION_LOST");
        break;
      case 6:
        Serial.println("WL_DISCONNECTED");
        break;
      default:
        Serial.println("_UNKNOWN");
        break;
   }
}

/*-----------------------------------------------------------------*/
/*  Starts everything related to WiFi and Internet connections.    */
/*  Variables are hard coded at the top of the software.           */ 
/*-----------------------------------------------------------------*/
int startWiFi(void)
{
  int count, j;

  Serial.println("** WiFi setup started");
  if( DEBUG )
  {
    Serial.print("**  Connecting to "); Serial.println(ssid); 
  }
  WiFi.hostname("ESP_IKEAHACK"); // Need to be unique.
  WiFi.begin(ssid, password);
  WiFi.mode(WIFI_STA);

  // Wait for connection
  count = 0;
  j = 1;
  while (WiFi.status() != WL_CONNECTED) // Wait for connection
  {
    strip.setPixelColor(j, strip.Color(0, 0, 128) ); strip.show(); // Visual indication for WiFi
    if( DEBUG ) Serial.print(".");    
    delay(500);
    count++;
    j++; 
    if( j == 7 ) j = 1;
    if( count > 40 ) // Adjust if it is waiting too long. Time = 500 * 40 usec.
    { 
      // Waited to long, no WiFi available
      Serial.println("\n!! No WiFi available - leaving WiFi init.");
      // Should enter some kind of visual error state.
      // Depends on availability of error LED.
      return false;
    }
  }
  // Clear LEDS
  for( j = 0; j < strip.numPixels(); j++) 
  {
    strip.setPixelColor(j, strip.Color(0, 0, 0) );
    strip.show();
  }
  if( count < 40 )
  { 
    // WiFi connection success.
    if( DEBUG )
    {
      Serial.println("");
      Serial.print("**  Connected to "); Serial.println(ssid);
      Serial.print("**  IP address: ");  Serial.println(WiFi.localIP());
      Serial.print("**  RSSI: ");        Serial.println(WiFi.RSSI());
      Serial.print("**  WiFi.status: "); showWiFiStatus( WiFi.status() );
    }
	// Associate the handler function to the path.
    server.on("/", handleRoot);        
	server.on("/NEOPixel_new_20191222_load.js", NEOPixel_load_script); 
    server.on("/API",handleAPI);       
    server.onNotFound(handleNotFound); // Any thing else
        
    server.begin();
    Serial.println("**  HTTP server started");
    // Print the IP address
    Serial.print("**   Use this URL : "); Serial.print("http://"); Serial.print(WiFi.localIP()); Serial.println("/");
    Serial.println("** WiFi setup done\n");    
  }
  return true;
}

/*----|| FUNCTIONS-end ||----------------------------------------*/


/*----|| SETUP ||--------------------------------------------*/
void setup() 
{
  uint16_t i;
  int j;

  // Init serial communication ------------------------------
  Serial.begin(9600); // Change here if you need faster communication
  delay(10);
  Serial.println( version ); 
  // Init serial communication END --------------------------
 
  Serial.println("** Setup started");
  
  // Init or read EEPROM settings. --------------------------
  Serial.println("** EEPROM Setup started");  
  EEPROM.begin(512); // Needed in order to start read or write to flash.
  
  // See if any old settings are saved (Addr 0 contains 72).
  if( EEPROM.read( 0 ) != 72 )
  {  // Save default configuration.

     EEPROM.write(0,            72); // From now on new values will be saved.
     EEPROM.write(L_STATUS,     lampStatus);
     EEPROM.write(C_RED,        lampRed);
     EEPROM.write(C_GREEN,      lampGreen);
     EEPROM.write(C_BLUE,       lampBlue);
     EEPROM.write(BRIGHTNESS,   lampBrightness);
     EEPROM.write(DELAY,        lampDelay);
     EEPROM.write(RAINBOWCYCLE, lampRainbowCycle);     
     EEPROM.write(RAINBOW,      lampRainbow);
     EEPROM.write(STATIC,       lampStatic);
     EEPROM.write(FADE,         lampFade);
     EEPROM.write(CANDLE,       lampCandle);
     // Store default SSID
	   for( j = 0; ((j < SSID_LEN) && (ssid[j] != 0x00)); j++ )
     {
       EEPROM.write(SSID_POS+j, ssid[j]);
     }
     EEPROM.write(SSID_POS+j, 0x00);

     // Store default password
     for( j = 0; ((j < PASSWORD_LEN) && (password[j] != 0x00)); j++ )
     {
       EEPROM.write(PASSWORD_POS+j, password[j]);
     }
     EEPROM.write(PASSWORD_POS+j, 0x00);
     EEPROM.commit(); // Needed in order to save content to flash.
  }
  
  // load saved configuration.
  lampStatus       = EEPROM.read(L_STATUS);
  lampRed          = EEPROM.read(C_RED);
  lampGreen        = EEPROM.read(C_GREEN);
  lampBlue         = EEPROM.read(C_BLUE);
  lampBrightness   = EEPROM.read(BRIGHTNESS);
  lampDelay        = EEPROM.read(DELAY);
  lampRainbowCycle = EEPROM.read(RAINBOWCYCLE);
  lampRainbow      = EEPROM.read(RAINBOW);
  lampStatic       = EEPROM.read(STATIC);
  lampFade         = EEPROM.read(FADE);
  lampCandle       = EEPROM.read(CANDLE);
  if( DEBUG ) 
  {
    Serial.print("**  Lamp status loaded: "); Serial.println(lampStatus);
    char  text[300];
    sprintf(text, "**  Colours loaded: %d, %d, %d", lampRed, lampGreen, lampBlue); 
    Serial.println(text);
    Serial.print("**  Lamp brightness loaded: "); Serial.println(lampBrightness);
    Serial.print("**  Lamp delay loaded: "); Serial.println(lampDelay);      
    if( lampRainbowCycle ) Serial.println("**  Rainbow cycle selected.");
    if( lampRainbow ) Serial.println("**  Rainbow selected.");
    if( lampStatic ) Serial.println("**  Static colour selected.");
    if( lampFade ) Serial.println("**  Fade selected.");
    if( lampCandle ) Serial.println("**  Candle selected.");
  }
  	 
  // Load stored SSID ----------------------------
  j = -1;
  do
  {
    j++;
    ssid[j] = EEPROM.read(SSID_POS+j);
  }
  while( (j < SSID_LEN) && (ssid[j] != 0x00) );
  ssid[j] = 0x00;
  if( DEBUG ) 
  { 
    Serial.print("**  SSID loaded: "); Serial.println( ssid);
  }
  // Load stored SSID END ------------------------
	
  // Load stored password ------------------------    
  j = -1;
  do
  {
    j++;
    password[j] = EEPROM.read(PASSWORD_POS+j);
  }
  while( (j < PASSWORD_LEN) && (password[j] != 0x00) );
  password[j] = 0x00;
  if( DEBUG ) 
  { 
    Serial.print("**  Password loaded: "); Serial.println( password );
  }
  // Load stored password END ---------------------
  Serial.println( "** EEPROM config loaded.");
  // Init or read EEPROM settings. END ----------------------
  
  // Init LED Strip -----------------------------------------
  strip.begin();
  // Turn LEDS off until SETUP done.
  for( i = 0; i < strip.numPixels(); i++) 
    strip.setPixelColor(i, strip.Color(0, 0, 0) ); strip.show();

  strip.setBrightness(lampBrightness); 
  strip.show();  
  Serial.println("** NEOpixel configured.");  
  // Init LED Strip END -------------------------------------

   // Check if user wants to change SSID and/or PASSWORD for network.
  Serial.println(">> Press enter to change SSID and/or password.");
  count = 0;
  j = 1;
  // Wait for any activity at the serial port
  while (Serial.available() <= 0) 
  {
    strip.setPixelColor(j, strip.Color(128, 0, 0) ); strip.show(); // Visual indication for user interaction
    Serial.print('.');   // send a capital . to show count down
    delay(500);
    strip.setPixelColor(j, strip.Color(0, 0, 0) ); strip.show(); // Visual indication for user interaction
    count++;
    j++;
    if( j == 7 ) j = 1;    
    if( count > 10 )
    { 
      // No activity at the serial port => EXIT
      Serial.print("\n>> Stopped waiting for SSID/Password setup.\n");
      break;
    }
  }
  
  // Clear LEDS
  for( i = 0; i < strip.numPixels(); i++) 
  {
    strip.setPixelColor(i, strip.Color(0, 0, 0) );
    strip.show();
  }  
  if( count < 10 ) // If too short increase the value. Time = 500 * 10 usec
  { 
    // Enter setup
    Serial.print("\nEntering setup mode.\n");
    Serial.print("Present SSID:     "); Serial.print(ssid);Serial.println("");
    Serial.print("Present PASSWORD: "); Serial.print(password);Serial.println("");

    // Clear incoming serial buffer before we continue.
    while( Serial.available() )
    { 
      // Are there anything to read?
      char getData = Serial.read();
    }     
    setupNotDone = true;
    while (setupNotDone)
    {
      cont   = true;
      inData = "";
      Serial.print("Valid commands: ssid, pssw, list, done.\n");  
      
      while( cont == true ) // look for commands
      {
        if (Serial.available() > 0) 
        {
          // get incoming byte:
          inByte  = Serial.read();
          inData += (char)inByte; 
          if (inByte == '\n') // command terminate char.
          {
            cont = false;  // A command has been received, process it and exit.
            if( inData == "ssid\n" ) // ssid command received
            {
              Serial.println("SSID:");              
              Serial.print("  Enter new SSID string (must be less than 20 char)\n");
              ByteCount = 0;
              
              memset(Buffer, 0, sizeof(Buffer));   // Clear contents of Buffer
              
              Serial.setTimeout(20000); // Wait for serial input for 20 sec.
              ByteCount = Serial.readBytesUntil('\n', Buffer, bSize);  // Receive new SSID.
              
              if (ByteCount > 0)  
              { 
                // Data received
                Serial.print("  New SSID: "); Serial.println(Buffer);
                if( ByteCount < 21 )
                { 
                  // New SSID has a valid length
                  Buffer[ByteCount] = 0x00; // Terminate new SSID
                  Serial.println("  Save new SSID (y/n)");
                  char  yesBuffer[2];   // Serial buffer for reply
                  if ( Serial.readBytesUntil('\n', yesBuffer, 2) > 0)
                  { 
                    // get incoming byte:
                    if( ((String)yesBuffer[0] == "y") || ((String)yesBuffer[0] == "Y"))
                    {
                      // Save new SSID 
                      memcpy(ssid, Buffer, ByteCount+1);
                      for( j = 0; ((j < SSID_LEN) && (Buffer[j] != 0x00)); j++ )
                      {
                        EEPROM.write(SSID_POS+j, ssid[j]);
                      }
                      EEPROM.write(SSID_POS+j, 0x00);
                      EEPROM.commit(); // Needed in order to save content to flash.
                      Serial.println("  New SSID saved.");
                    }
                    else
                      Serial.println("  New SSID not saved.");
                  }
                  else
                    Serial.println("  Timeout. SSID not saved.");
                }
                else
                  Serial.println("  New SSID too long. Please try again.");
              }
              else
                Serial.println("  Please try again");

              memset(Buffer, 0, sizeof(Buffer));   // Clear contents of Buffer
            } // END ssid command received 

            if( inData == "pssw\n" ) // password command received
            {
              Serial.println("PSSW:");
              Serial.print("  Enter new password string (must be less than 20 char)\n");
              ByteCount = 0;
              
              memset(Buffer, 0, sizeof(Buffer));   // Clear contents of Buffer
              
              Serial.setTimeout(20000); // Wait for serial input for 20 sec.
              ByteCount = Serial.readBytesUntil('\n', Buffer, bSize);  // Receive new SSID.
              
              if (ByteCount > 0)  
              { 
                // Data received
                Serial.print("  New password: "); Serial.println(Buffer);
                if( ByteCount < 21 )
                { 
                  // New password has a valid length
                  Buffer[ByteCount] = 0x00; // Terminate new password
                  Serial.println("  Save new password (y/n)");
                  char yesBuffer[2];   // Serial buffer for reply
                  if ( Serial.readBytesUntil('\n', yesBuffer, 2) > 0)
                  { 
                    // get incoming byte:
                    if( ((String)yesBuffer[0] == "y") || ((String)yesBuffer[0] == "Y"))
                    {
                      // Save new password 
                      memcpy(password, Buffer, ByteCount+1);
                      for( j = 0; ((j < PASSWORD_LEN) && (Buffer[j] != 0x00)); j++ )
                      {
                        EEPROM.write(PASSWORD_POS+j, password[j]);
                      }
                      EEPROM.write(PASSWORD_POS+j, 0x00);
                      EEPROM.commit(); // Needed in order to save content to flash.
                      Serial.println("  New password saved.");
                    }
                    else
                      Serial.println("  New password not saved.");
                  }
                  else
                    Serial.println("  Timeout. Password not saved.");
                }
                else
                  Serial.println("  New password too long. Please try again.");
              }
              else
                Serial.println("  Please try again");

              memset(Buffer, 0, sizeof(Buffer));   // Clear contents of Buffer               
            } // END list command received
                        
            if( inData == "list\n" ) // list command received
            {
               Serial.print("LIST:\n");
               Serial.print("  Present SSID:     "); Serial.print(ssid);Serial.println("");
               Serial.print("  Present PASSWORD: "); Serial.print(password);Serial.println("");
               memset(Buffer, 0, sizeof(Buffer));   // Clear contents of Buffer
            } // END list command received
            
            if( inData == "done\n" ) // done command received
            {
              Serial.println("DONE:");

              memset(Buffer, 0, sizeof(Buffer)); // Clear contents of Buffer              
              setupNotDone = false; // Exit setup and launch lamp.
            } // END done command received
          } // END if (inByte == '\n')
        } // END if (Serial.available() > 0)
      } // END look for commands
    }// END while( cont )
  }

  // 
  if( !startWiFi() ) 
  {
    // enter error state: NO WiFi.
    // We need a WiFI connection to control the device
	  // If it is a standalone device uncomment the next line.
    errorState( 1 );
  }
  // Setup finished carry on
  
  if( !startSPIFFS() )
  {
    // enter error state: NO WiFi.
    // We need a WiFI connection to control the device
    errorState( 1 );
  }  
  
  Serial.println("Entering Lamp mode.");  
}
/*----|| SETUP-end ||----------------------------------------*/

/*----|| LOOP ||--------------------------------------------*/
void loop() 
{
  uint16_t i;
  server.handleClient(); // Check for incoming requests
  if( lampStatus )
  {
    if( lampRainbow )
      rainbow(lampDelay);
    else if( lampRainbowCycle )
      rainbowCycle(lampDelay);
    else if( lampFade )
      fade(lampDelay);
    else if( lampCandle )
      candle(lampDelay);
    else if ( lampStatic)
    {
      for( i = 0; i < strip.numPixels(); i++) 
      {
        strip.setPixelColor(i, strip.Color(lampRed, lampGreen, lampBlue) );
        strip.show();
      }
    }
  }
}
/*----|| LOOP-end ||----------------------------------------*/
